(ns clj-ml1.lmatrix
  "Laplacian matrix example from Ch. 1"
  (:use clojure.core.matrix
        [incanter.charts :only [xy-plot add-points]]
        [incanter.core   :only [view]])
  (:require [clojure.core.matrix.operators :as M]
            [clatrix.core :as cl]))

(defn lmatrix [n]
  (compute-matrix :clatrix [n (+ n 2)]
                  (fn [i j] ({0 -1 , 1 2, 2 -1} (- j i) 0))))

;; user> (pm (lmatrix 4))
;; [[-1.000  2.000 -1.000  0.000  0.000  0.000]
;;  [ 0.000 -1.000  2.000 -1.000  0.000  0.000]
;;  [ 0.000  0.000 -1.000  2.000 -1.000  0.000]
;;  [ 0.000  0.000  0.000 -1.000  2.000 -1.000]]
;; nil

(defn problem
  "Return a map of the problem setup for a given matrix size,
   number of observed values and regularization parameter"
  [n n-observed lambda]
  (let [i (shuffle (range n))]
    {:L        (M/* (lmatrix n) lambda)
     :observed (take n-observed i)
     :hidden   (drop n-observed i)
     :observed-values (matrix :clatrix
                              (repeatedly n-observed rand))}))

(defn solve
  "Return a map containing the approximated value y of each hidden point x"
  [{:keys [L observed hidden observed-values] :as problem}]
  (let [nc  (column-count L)
        nr  (row-count L)
        L1  (cl/get L (range nr) hidden)
        L2  (cl/get L (range nr) observed)
        l11 (M/* (transpose L1) L1)
        l12 (M/* (transpose L1) L2)]
    (assoc problem :hidden-values
      (M/* -1 (inverse l11) l12 observed-values))))

(defn plot-points
  "Plots sample points of a solution s"
  [s]
  (let [X (concat (:hidden s) (:observed s))
        Y (concat (:hidden-values s) (:observed-values s))]
    (view
     (add-points
      (xy-plot X Y) (:observed s) (:observed-values s)))))

(defn plot-rand-sample []
  (plot-points (solve (problem 150 10 30))))

;; run using (plot-rand-sample)
