(ns clj-ml3.bayes-implementation)

;; A Naive Bayesian Classifier

;; sea bass are mostly long and light in color
(defn make-sea-bass []
  #{:sea-bass
    (if (< (rand) 0.2) :fat :thin)
    (if (< (rand) 0.7) :long :short)
    (if (< (rand) 0.8) :light :dark)})

;; salmon are mostly fat and dark
(defn make-salmon []
  #{:salmon
    (if (< (rand) 0.8) :fat :thin)
    (if (< (rand) 0.5) :long :short)
    (if (< (rand) 0.3) :light :dark)})

(defn make-sample-fish []
  (if (< (rand) 0.3) (make-sea-bass) (make-salmon)))

(def fish-training-data
  (for [i (range 10000)] (make-sample-fish)))

(defn probability
  "Calculates the probability of a specific category
   given some attributes, depending on the training data."
  [attribute & {:keys
                [category prior-positive prior-negative data]
                :or {category nil
                     data fish-training-data}}]
  (let [by-category (if category
                      (filter category data)
                      data)
        positive (count (filter attribute by-category))
        negative (- (count by-category) positive)
        total (+ positive negative)]
    (/ positive negative)))

;; user> (probability :dark :category :salmon)
;; 1204/1733
;; user> (probability :dark :category :sea-bass)
;; 621/3068
;; user> (probability :light :category :salmon)
;; 529/1733
;; user> (probability :light :category :sea-bass)
;; 2447/3068

(defn evidence-of-sea-bass [& attrs]
  (let [attr-probs (map #(probability % :category :sea-bass) attrs)
        class-and-attr-prob (conj attr-probs
                                  (probability :sea-bass))]
    (float (apply * class-and-attr-prob))))

(defn evidence-of-salmon [& attrs]
  (let [attr-probs (map #(probability % :category :salmon) attrs)
        class-and-attr-prob (conj attr-probs
                                  (probability :salmon))]
    (float (apply * class-and-attr-prob))))

;; user> (evidence-of-salmon :dark)
;; 0.4816
;; user> (evidence-of-salmon :dark :long)
;; 0.2396884
;; user> (evidence-of-salmon)
;; 0.6932
;;
;; user> (+ (evidence-of-sea-bass) (evidence-of-salmon))
;; 1.0000000298023224

;; generalized as follows
(defn evidence-of-category-with-attrs
  [category & attrs]
  (let [attr-probs (map #(probability % :category category) attrs)
        class-and-attr-prob (conj attr-probs
                                  (probability category))]
    (float (apply * class-and-attr-prob))))

;; user> (evidence-of-salmon :dark :fat)
;; 0.38502988
;; user> (evidence-of-category-with-attrs :salmon :dark :fat)
;; 0.38502988

(def probability-dark-long-fat-is-salmon
  (let [attrs [:dark :long :fat]
        sea-bass? (apply evidence-of-sea-bass attrs)
        salmon? (apply evidence-of-salmon attrs)]
    (/ salmon?
       (+ sea-bass? salmon?))))

;; user> probability-dark-long-fat-is-salmon
;; 0.957091799207812

(def probability-dark-long-fat-is-sea-bass
  (let [attrs [:dark :short :fat]
        sea-bass? (apply evidence-of-sea-bass attrs)
        salmon? (apply evidence-of-salmon attrs)]
    (/ sea-bass?
       (+ sea-bass? salmon?))))

;; generalizing this,
;;

(defn make-category-probability-pair
  [category attrs]
  (let [evidence-of-category (apply evidence-of-category-with-attrs
                                    category attrs)]
    {:category category
     :evidence evidence-of-category}))

(defn calculate-probability-of-category
  [sum-of-evidences pair]
  (let [probability-of-category (/ (:evidence pair)
                                   sum-of-evidences)]
    (assoc pair
      :probability probability-of-category)))

(defn classify-by-attrs
  "Performs Bayesian classification of the attributes, given some categories.
   Returns a map containing the predicted category and the category's
   probability of occurence."
  [categories & attrs]
  (let [pairs (map #(make-category-probability-pair % attrs)
                   categories)
        sum-of-evidences (reduce + (map :evidence pairs))
        probabilities (map #(calculate-probability-of-category
                            sum-of-evidences %)
                          pairs)
        sorted-probabilities (sort-by :probability probabilities)
        predicted-category (last sorted-probabilities)]
    predicted-category))

;; user> (classify-by-attrs [:salmon :sea-bass] :dark :long :fat)
;; {:probability 0.957091799207812, :category :salmon, :evidence 0.1949689}
;;
;; user> (classify-by-attrs [:salmon :sea-bass] :dark)
;; {:probability 0.8857825967670728, :category :salmon, :evidence 0.4816}
;; user> (classify-by-attrs [:salmon :sea-bass] :thin)
;; {:probability 0.6369809383442954, :category :sea-bass, :evidence 0.2439}
;; user> (classify-by-attrs [:salmon :sea-bass] :light)
;; {:probability 0.5362699908806723, :category :sea-bass, :evidence 0.2447}
;; user> (classify-by-attrs [:salmon] :dark)
;;
;; {:probability 1.0, :category :salmon, :evidence 0.4816}
;; user> (classify-by-attrs [:salmon])
;; {:probability 1.0, :category :salmon, :evidence 0.6932}
